/***************************************************************************
 Ice Tube Clock firmware August 13, 2009
 (c) 2009 Limor Fried / Adafruit Industries

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
****************************************************************************/

#include <avr/io.h>      // this contains all the IO port definitions
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <string.h>
#include <avr/wdt.h> 

#include "iv.h"
#include "util.h"

#define REGION_US 0
#define REGION_EU 1
uint8_t region = REGION_US;

volatile uint8_t time_s, time_m, time_h;
// MEME
volatile uint8_t timeunknown = 0;

volatile uint8_t date_m, date_d, date_y;
volatile uint8_t alarming, alarm_on, alarm_h, alarm_m;
volatile uint8_t displaymode;
volatile uint8_t volume;
volatile uint8_t sleepmode = 0;


#define DISPLAYSIZE 9
uint8_t display[DISPLAYSIZE]; // stores segments, not values!
uint8_t currdigit = 0;

uint8_t prev_display[DISPLAYSIZE]; // fading from
uint8_t fades[DISPLAYSIZE];

const uint8_t alphatable[] PROGMEM = {
	0xFA, /* a */
	0x3E, /* b */
	0x1A, /* c */
	0x7A, /* d */
	0xDE, /* e */
	0x8E, /* f */
	0xF6, /* g */
	0x2E, /* h */
	0x60, /* i */
	0x78, /* j */
	0xAE, //k
	0x1C, // l
	0xAA, // m
	0x2A, // n
	0x3A, // o
	0xCE, //p
	0xF3, // q
	0x0A, //r
	0xB6, //s
	0x1E, //t
	0x38, //u
	0x38, //v // fix?
	0xB8, //w
	0x6E, //x
	0x76, // y
	0xDA, //z
	/* more */
};
PGM_P alphatable_p PROGMEM = alphatable;

const uint8_t numbertable[] PROGMEM = { 
  0xFC /* 0 */, 
  0x60 /* 1 */,
  0xDA /* 2 */,
  0xF2 /* 3 */,
  0x66 /* 4 */,
  0xB6 /* 5 */,
  0xBE, /* 6 */
  0xE0, /* 7 */
  0xFE, /* 8 */
  0xE6, /* 9 */
};
PGM_P numbertable_p PROGMEM = numbertable;

const uint8_t digittable[] PROGMEM = {
  DIG_9, DIG_8, DIG_7, DIG_6, DIG_5, DIG_4, DIG_3, DIG_2, DIG_1
};
PGM_P digittable_p PROGMEM = digittable;


const uint8_t segmenttable[] PROGMEM = {
  SEG_H, SEG_G,  SEG_F,  SEG_E,  SEG_D,  SEG_C,  SEG_B,  SEG_A 
};
PGM_P segmenttable_p PROGMEM = segmenttable;

uint16_t muxdiv = 0;
#define MUX_DIVIDER (300 / DISPLAYSIZE)
uint16_t alarmdiv = 0;
#define ALARM_DIVIDER 100
#define MAXSNOOZE 600 // 10 minutes
uint16_t snoozetimer = 0;

volatile uint16_t milliseconds = 0;
void delayms(uint16_t ms) {
  uint16_t wait;
  sei();
  wait = milliseconds;
  while (milliseconds < (wait+ms));
}

void setsnooze(void) {
  //snoozetimer = eeprom_read_byte((uint8_t *)EE_SNOOZE);
  //snoozetimer *= 60; // convert minutes to seconds
  snoozetimer = MAXSNOOZE;
  DEBUGP("snooze");
  display_str("snoozing");
  displaymode = SHOW_SNOOZE;
  delayms(1000);
  displaymode = SHOW_TIME;
}

// called @ (F_CPU/256) = ~30khz (31.25 khz)
SIGNAL (SIG_OUTPUT_COMPARE0A) {
  uint8_t d;
  sei();

  wdt_reset();

  muxdiv++;
  if (muxdiv < MUX_DIVIDER)
    return;
  muxdiv = 0;
  // now at 100hz

  milliseconds++;

  if (currdigit >= DISPLAYSIZE)
    currdigit = 0;

  d = display[currdigit];

  /*
    uint8_t f = fades[currdigit];
  
  if ((prev_display[currdigit] != display[currdigit]) && (f == 0)) {
    uart_putw_dec(currdigit);
    fades[currdigit] = 6;  // 10 millisecond fade, approx
  }

  if (f < 3) {
    d = prev_display[currdigit];
  } 
  if (f!=0) {
    fades[currdigit]--;
    if (fades[currdigit] == 0) {
      prev_display[currdigit] = display[currdigit];
    }
    }*/ 
  setdisplay(currdigit, d);
  currdigit++;

  // check if we should have the alarm on
  if (alarming && !snoozetimer) {
    alarmdiv++;
    if (alarmdiv > ALARM_DIVIDER) {
      alarmdiv = 0;
    } else {
      return;
    }
    ICR1 = 250;
    OCR1A = OCR1B = ICR1/2;
    // ok alarm is ringing!'
    if (alarming & 0xF0) { // top bit indicates pulsing alarm state
      alarming &= ~0xF0;
      TCCR1B &= ~_BV(CS11); // turn it off!
    } else {
      alarming |= 0xF0;
      TCCR1B |= _BV(CS11); // turn it on!
    }
  }
  
}


volatile uint8_t last_buttonstate = 0, just_pressed = 0, pressed = 0;
volatile uint8_t buttonholdcounter = 0;

SIGNAL(SIG_PIN_CHANGE2) {
  sei();
  if (! (PIND & _BV(BUTTON1))) {
    // button1 is pressed
    if (! (last_buttonstate & 0x1)) { // was not pressed before
      delayms(10);
      if (PIND & _BV(BUTTON1))       // filter out bounces
	return;
      tick();
      // check if we will snag this button press for snoozing
      if (alarming) {
	// turn on snooze
	setsnooze();
	return;
      }
      last_buttonstate |= 0x1;
      just_pressed |= 0x1;
      //DEBUGP("b1");
    }
  } else {
    last_buttonstate &= ~0x1;
  }
  if (! (PIND & _BV(BUTTON3))) {
    // button3 is pressed
    if (! (last_buttonstate & 0x4)) { // was not pressed before
      delayms(10);
      if (PIND & _BV(BUTTON3))
	return;
      buttonholdcounter = 2;
      while (buttonholdcounter) {
	if (PIND & _BV(BUTTON3)) { // released
	  tick();
	  last_buttonstate &= ~0x4;
	  // check if we will snag this button press for snoozing
	  if (alarming) {
	    // turn on snooze
	    setsnooze();
	    return;
	  }
	  DEBUGP("b3");
	  just_pressed |= 0x4;
	  return;
	}
      }
      last_buttonstate |= 0x4;
      // held down 
      pressed |= 0x4;
    }
  } else {
    pressed = 0;
    last_buttonstate &= ~0x4;
  }
}

SIGNAL(SIG_PIN_CHANGE0) {
  sei();
  if (! (PINB & _BV(BUTTON2))) {
    // button2 is pressed
    if (! (last_buttonstate & 0x2)) { // was not pressed before
      delayms(10);
      if (PINB & _BV(BUTTON2))
	return;
      tick();
      // check if we will snag this button press for snoozing
      if (alarming) {
	// turn on snooze
	setsnooze();
	return;
      }
      last_buttonstate |= 0x2;
      just_pressed |= 0x2;
      DEBUGP("b2");
    }
  } else {
    last_buttonstate &= ~0x2;
  }
}

#define INACTIVITYTIMEOUT 10
volatile uint8_t timeoutcounter = 0;


SIGNAL (TIMER2_OVF_vect) {
  CLKPR = _BV(CLKPCE); 
  CLKPR = 0;
  time_s++;

  if (time_s >= 60) {
    time_s = 0;
    time_m++;
  }

  if (time_m >= 60) {
    time_m = 0;
  // write the time to the EEPROM
    time_h++; 
    eeprom_write_byte((uint8_t *)EE_HOUR, time_h);
    eeprom_write_byte((uint8_t *)EE_MIN, time_m);
  }
  if (time_h >= 24) {
    time_h = 0;

    date_d++;
    eeprom_write_byte((uint8_t *)EE_DAY, date_d);
  }

#if (DEBUG == 1)
  if (! sleepmode) {
    uart_putw_dec(time_h);
    uart_putchar(':');
    uart_putw_dec(time_m);
    uart_putchar(':');
    uart_putw_dec(time_s);
    putstring_nl("");
  }
  #endif

  if ((date_d > 31) ||
      ((date_d == 31) && ((date_m == 4)||(date_m == 6)||(date_m == 9)||(date_m == 11))) ||
      ((date_d == 30) && (date_m == 2)) ||
      ((date_d == 29) && (date_m == 2) && !leapyear(2000+date_y))) {
    date_d = 1;
    date_m++;
    eeprom_write_byte((uint8_t *)EE_MONTH, date_m);
  }
  
  if (date_m >= 13) {
    date_y++;
    date_m = 1;
    eeprom_write_byte((uint8_t *)EE_YEAR, date_y);
  }
  
  if (sleepmode)
    return;
   
  if (displaymode == SHOW_TIME) {
    if (timeunknown && (time_s % 2)) {
      display_str("        ");
    } else {
      display_time(time_h, time_m, time_s);
    }
    if (alarm_on)
      display[0] |= 0x2;
    else 
      display[0] &= ~0x2;
    
  }
  if (alarm_on && (alarm_h == time_h) && (alarm_m == time_m) && (time_s == 0)) {
    DEBUGP("alarm on!");
    alarming = 1;
    snoozetimer = 0;
  }

  if (timeoutcounter)
    timeoutcounter--;
  if (buttonholdcounter)
    buttonholdcounter--;
  if (snoozetimer) {
    snoozetimer--;
    if (snoozetimer % 2) 
      display[0] |= 0x2;
    else
      display[0] &= ~0x2;
  }
}

SIGNAL(SIG_INTERRUPT0) {
  uart_putchar('i');
  uint8_t x = ALARM_PIN & _BV(ALARM);
  sei();
  delayms(10); // wait for debouncing
  if (x != (ALARM_PIN & _BV(ALARM)))
    return;
  setalarmstate();
}



SIGNAL(SIG_COMPARATOR) {
  //DEBUGP("COMP");
  if (ACSR & _BV(ACO)) {
    //DEBUGP("HIGH");
    if (!sleepmode) {
      VFDSWITCH_PORT |= _BV(VFDSWITCH); // turn off display
      VFDCLK_PORT &= ~_BV(VFDCLK) & ~_BV(VFDDATA); // no power to vfdchip
      BOOST_PORT &= ~_BV(BOOST); // pull boost fet low
      SPCR  &= ~_BV(SPE); // turn off spi
      eeprom_write_byte((uint8_t *)EE_MIN, time_m);
      eeprom_write_byte((uint8_t *)EE_SEC, time_s);
      DEBUGP("z");
      TCCR0B = 0; // no boost
      volume = 0; // low power buzzer
      PCICR = 0;  // ignore buttons

      app_start();
    }
  } else {
    //DEBUGP("LOW");
    if (sleepmode) {
      eeprom_write_byte((uint8_t *)EE_MIN, time_m);
      eeprom_write_byte((uint8_t *)EE_SEC, time_s);
      DEBUGP("WAKERESET"); 
      app_start();
    }
  }
}
/*********************** Main app **********/

uint32_t t;

void gotosleep(void) {
  // battery
  //if (sleepmode) //already asleep?
  //  return;
  //DEBUGP("sleeptime");
  
  sleepmode = 1;
  VFDSWITCH_PORT |= _BV(VFDSWITCH); // turn off display
  SPCR  &= ~_BV(SPE); // turn off spi
  VFDCLK_PORT &= ~_BV(VFDCLK) & ~_BV(VFDDATA); // no power to vfdchip
  BOOST_PORT &= ~_BV(BOOST); // pull boost fet low
  TCCR0B = 0; // no boost
  volume = 0; // low power buzzer
  PCICR = 0;  // ignore buttons

  // sleep time!
  //beep(3520, 1);
  //beep(1760, 1);
  //beep(880, 1);
  // turn beeper off
  PORTB &= ~_BV(SPK1) & ~_BV(SPK2); 
  
  // turn off pullups
  PORTD &= ~_BV(BUTTON1) & ~_BV(BUTTON3);
  PORTB &= ~_BV(BUTTON2);
  DDRD &= ~_BV(BUTTON1) & ~_BV(BUTTON3);
  DDRB &= ~_BV(BUTTON2);
  ALARM_PORT &= ~_BV(ALARM);
  ALARM_DDR &= ~_BV(ALARM);
  

  // reduce the clock speed
  CLKPR = _BV(CLKPCE);
  CLKPR = _BV(CLKPS3);
  
  //  PPR |= _BV(PRUSART0) | _BV(PRADC) | _BV(PRSPI) | _BV(PRTIM1) | _BV(PRTIM0) | _BV(PRTWI);
  PORTC |= _BV(4);  // sleep signal
  SMCR |= _BV(SM1) | _BV(SM0) | _BV(SE); // sleep mode
  asm("sleep"); 
  CLKPR = _BV(CLKPCE);
  CLKPR = 0;
  PORTC &= ~_BV(4);
}
 
 void wakeup(void) {
   if (!sleepmode)
     return;
   CLKPR = _BV(CLKPCE);
   CLKPR = 0;
   DEBUGP("waketime");
   sleepmode = 0;
   // plugged in
   // wait to verify
   _delay_ms(20);
   if (ACSR & _BV(ACO)) 
     return;
   
   // turn on pullups
   initbuttons();

   // turn on boost
   boost_init(eeprom_read_byte((uint8_t *)EE_BRIGHT));
   // turn on vfd control
   vfd_init();

   // turn on display
   VFDSWITCH_PORT &= ~_BV(VFDSWITCH); 
   VFDBLANK_PORT &= ~_BV(VFDBLANK);
   volume = eeprom_read_byte((uint8_t *)EE_VOLUME); // reset
   
   speaker_init();

   wdt_reset();

   setalarmstate();

   // wake up sound
   beep(880, 1);
   beep(1760, 1);
   beep(3520, 1);

   wdt_reset();
 }


void initbuttons(void) {
    DDRB =  _BV(VFDCLK) | _BV(VFDDATA) | _BV(SPK1) | _BV(SPK2);
    DDRD = _BV(BOOST) | _BV(VFDSWITCH);
    DDRC = _BV(VFDLOAD) | _BV(VFDBLANK) | _BV(4);
    PORTD = _BV(BUTTON1) | _BV(BUTTON3) | _BV(ALARM);
    PORTB = _BV(BUTTON2);

    PCICR = _BV(PCIE0) | _BV(PCIE2);
    PCMSK0 = _BV(PCINT0);
    PCMSK2 = _BV(PCINT21) | _BV(PCINT20);
    

}



int main(void) {
  //  uint8_t i;
  uint8_t mcustate;

  // check if we were reset
  mcustate = MCUSR;
  MCUSR = 0;

  wdt_disable();
  
  //WDTCSR |= _BV(WDCE) | _BV(WDE); // turn off watchdog timer!
  //WDTCSR = 0;

  // now turn it back on... 2 second time out
  //WDTCSR |= _BV(WDP0) | _BV(WDP1) | _BV(WDP2);
  //WDTCSR = _BV(WDE);
  wdt_enable(WDTO_2S);
  wdt_reset();

  // setup uart
  uart_init(BRRL_192);
  //DEBUGP("VFD Clock");
  DEBUGP("!");

  //DEBUGP("turning on anacomp");
  // set up analog comparator
  ACSR = _BV(ACBG) | _BV(ACIE); // use bandgap, intr. on toggle!
  _delay_ms(1);
  // settle!
  if (ACSR & _BV(ACO)) {
    // hmm we should not interrupt here
    ACSR |= _BV(ACI);

    // even in low power mode, we run the clock 
    DEBUGP("clock init");
    clock_init();  

  } else {
    // we aren't in low power mode so init stuff

    // init io's
    initbuttons();
    
    VFDSWITCH_PORT &= ~_BV(VFDSWITCH);
    
    DEBUGP("turning on buttons");
    // set up button interrupts
    DEBUGP("turning on alarmsw");
    // set off an interrupt if alarm is set or unset
    EICRA = _BV(ISC00);
    EIMSK = _BV(INT0);
  
    displaymode = SHOW_TIME;
    DEBUGP("vfd init");
    vfd_init();
    
    DEBUGP("boost init");
    boost_init(eeprom_read_byte((uint8_t *)EE_BRIGHT));
    sei();

    region = eeprom_read_byte((uint8_t *)EE_REGION);
    
    DEBUGP("speaker init");
    speaker_init();

    DEBUGP("clock init");
    clock_init();  

    DEBUGP("alarm init");
    setalarmstate();
 }

  while (1) {
    //_delay_ms(100);
    wdt_reset();
    //uart_putc_hex(ACSR);
    if (ACSR & _BV(ACO)) {
      //DEBUGP("SLEEPYTIME");
      gotosleep();
      continue;
    }
    //DEBUGP(".");
    if (just_pressed & 0x1) {
      just_pressed = 0;
      switch(displaymode) {
      case (SHOW_TIME):
	displaymode = SET_ALARM;
	display_str("set alarm");
	set_alarm();
	timeunknown = 0;
	break;
      case (SET_ALARM):
	displaymode = SET_TIME;
	display_str("set time");
	set_time();
	break;
      case (SET_TIME):
	displaymode = SET_DATE;
	display_str("set date");
	set_date();
	break;
      case (SET_DATE):
	displaymode = SET_BRIGHTNESS;
	display_str("set brit");
	set_brightness();
	break;
      case (SET_BRIGHTNESS):
	displaymode = SET_VOLUME;
	display_str("set vol ");
	set_volume();
	break;
      case (SET_VOLUME):
	displaymode = SET_REGION;
	display_str("set regn");
	set_region();
	break;
	/*
      case (SET_REGION):
	displaymode = SET_SNOOZE;
	display_str("set snoz");
	set_snooze();
	break;
	*/
      default:
	displaymode = SHOW_TIME;
      }
    } else if (just_pressed & 0x2) {
      just_pressed = 0;
      displaymode = NONE;
      display_date(DAY);

      wdt_reset();
      delayms(1500);
      wdt_reset();

      displaymode = SHOW_TIME;     
    } else if (just_pressed & 0x4) {
      just_pressed = 0;
      displaymode = SHOW_ALARM;

      wdt_reset();
      delayms(1500);
      wdt_reset();

      displaymode = SHOW_TIME;     
    }
  }
}

/**************************** SUB-MENUS *****************************/
void set_region(void) {
  uint8_t mode = SHOW_MENU;

  timeoutcounter = INACTIVITYTIMEOUT;;  
  region = eeprom_read_byte((uint8_t *)EE_REGION);

  while (1) {
    if (just_pressed || pressed) {
      timeoutcounter = INACTIVITYTIMEOUT;;  
      // timeout w/no buttons pressed after 3 seconds?
    } else if (!timeoutcounter) {
      //timed out!
      displaymode = SHOW_TIME;     
      return;
    }
    if (just_pressed & 0x1) { // mode change
      return;
    }
    if (just_pressed & 0x2) {
      just_pressed = 0;
      if (mode == SHOW_MENU) {
	// start!
	mode = SET_REG;
	// display region
	if (region == REGION_US) {
	  display_str("usa-12hr");
	} else {
	  display_str("eur-24hr");
	}
      } else {	
	displaymode = SHOW_TIME;
	return;
      }
    }
    if (just_pressed & 0x4) {
      just_pressed = 0;
      if (mode == SET_REG) {
	region = !region;
	if (region == REGION_US) {
	  display_str("usa-12hr");
	} else {
	  display_str("eur-24hr");
	}
	eeprom_write_byte((uint8_t *)EE_REGION, region);
	speaker_init();
	beep(4000, 1);
      }
    }
  }
}

void set_volume(void) {
  uint8_t mode = SHOW_MENU;
  uint8_t volume;

  timeoutcounter = INACTIVITYTIMEOUT;;  
  volume = eeprom_read_byte((uint8_t *)EE_VOLUME);

  while (1) {
    if (just_pressed || pressed) {
      timeoutcounter = INACTIVITYTIMEOUT;;  
      // timeout w/no buttons pressed after 3 seconds?
    } else if (!timeoutcounter) {
      //timed out!
      displaymode = SHOW_TIME;     
      return;
    }
    if (just_pressed & 0x1) { // mode change
      return;
    }
    if (just_pressed & 0x2) {
      just_pressed = 0;
      if (mode == SHOW_MENU) {
	// start!
	mode = SET_VOL;
	// display volume
	if (volume) {
	  display_str("vol high");
	  display[5] |= 0x1;
	} else {
	  display_str("vol  low");
	}
	display[6] |= 0x1;
	display[7] |= 0x1;
	display[8] |= 0x1;
      } else {	
	displaymode = SHOW_TIME;
	return;
      }
    }
    if (just_pressed & 0x4) {
      just_pressed = 0;
      if (mode == SET_VOL) {
	volume = !volume;
	if (volume) {
	  display_str("vol high");
	  display[5] |= 0x1;
	} else {
	  display_str("vol  low");
	}
	display[6] |= 0x1;
	display[7] |= 0x1;
	display[8] |= 0x1;
	eeprom_write_byte((uint8_t *)EE_VOLUME, volume);
	speaker_init();
	beep(4000, 1);
      }
    }
  }
}

void set_brightness(void) {
  uint8_t mode = SHOW_MENU;
  uint8_t brightness;

  timeoutcounter = INACTIVITYTIMEOUT;;  
  brightness = eeprom_read_byte((uint8_t *)EE_BRIGHT);

  while (1) {
    if (just_pressed || pressed) {
      timeoutcounter = INACTIVITYTIMEOUT;;  
      // timeout w/no buttons pressed after 3 seconds?
    } else if (!timeoutcounter) {
      //timed out!
      displaymode = SHOW_TIME;     
      return;
    }
    if (just_pressed & 0x1) { // mode change
      return;
    }
    if (just_pressed & 0x2) {

      just_pressed = 0;
      if (mode == SHOW_MENU) {
	// start!
	mode = SET_BRITE;
	// display brightness
	display_str("brite ");
	display[7] = pgm_read_byte(numbertable_p + (brightness / 10)) | 0x1;
	display[8] = pgm_read_byte(numbertable_p + (brightness % 10)) | 0x1;
      } else {	
	displaymode = SHOW_TIME;
	return;
      }
    }
    if ((just_pressed & 0x4) || (pressed & 0x4)) {
      just_pressed = 0;
      if (mode == SET_BRITE) {
	brightness += 10;
	if (brightness >= 100)
	  brightness = 10;
	display[7] = pgm_read_byte(numbertable_p + (brightness / 10)) | 0x1;
	display[8] = pgm_read_byte(numbertable_p + (brightness % 10)) | 0x1;
	eeprom_write_byte((uint8_t *)EE_BRIGHT, brightness);
	OCR0A = brightness;
      }
    }
  }
}

/*
void set_snooze(void) {
  uint8_t mode = SHOW_MENU;
  uint8_t snooze;

  timeoutcounter = INACTIVITYTIMEOUT;;  
  snooze = eeprom_read_byte((uint8_t *)EE_SNOOZE);

  while (1) {
    if (just_pressed || pressed) {
      timeoutcounter = INACTIVITYTIMEOUT;;  
      // timeout w/no buttons pressed after 3 seconds?
    } else if (!timeoutcounter) {
      //timed out!
      displaymode = SHOW_TIME;     
      return;
    }
    if (just_pressed & 0x1) { // mode change
      return;
    }
    if (just_pressed & 0x2) {

      just_pressed = 0;
      if (mode == SHOW_MENU) {
	// start!
	mode = SET_SNOOZE;
	// display snooze
	display_str("   minut");
	display[1] = pgm_read_byte(numbertable_p + (snooze / 10)) | 0x1;
	display[2] = pgm_read_byte(numbertable_p + (snooze % 10)) | 0x1;
      } else { 
	displaymode = SHOW_TIME;
	return;
      }
    }
    if ((just_pressed & 0x4) || (pressed & 0x4)) {
      just_pressed = 0;
      if (mode == SET_SNOOZE) {
        snooze ++;
	if (snooze >= 100)
	  snooze = 0;
	display[1] = pgm_read_byte(numbertable_p + (snooze / 10)) | 0x1;
	display[2] = pgm_read_byte(numbertable_p + (snooze % 10)) | 0x1;
	eeprom_write_byte((uint8_t *)EE_SNOOZE, snooze);
      }

      if (pressed & 0x4)
	delayms(75);

    }
  }
}
*/

void set_date(void) {
  uint8_t mode = SHOW_MENU;

  timeoutcounter = INACTIVITYTIMEOUT;;  

  while (1) {
    if (just_pressed || pressed) {
      timeoutcounter = INACTIVITYTIMEOUT;;  
      // timeout w/no buttons pressed after 3 seconds?
    } else if (!timeoutcounter) {
      //timed out!
      displaymode = SHOW_TIME;     
      return;
    }
    if (just_pressed & 0x1) { // mode change
      return;
    }
    if (just_pressed & 0x2) {

      just_pressed = 0;
      if (mode == SHOW_MENU) {
	// start!
	if (region == REGION_US) {
	  mode = SET_MONTH;
	}
	else {
	  DEBUGP("Set day");
	  mode = SET_DAY;
	}
	display_date(DATE);
	display[1] |= 0x1;
	display[2] |= 0x1;
      } else if (((mode == SET_MONTH) && (region == REGION_US)) ||
		 ((mode == SET_DAY) && (region == REGION_EU))) {
	if (region == REGION_US)
	  mode = SET_DAY;
	else
	  mode = SET_MONTH;
	display_date(DATE);
	display[4] |= 0x1;
	display[5] |= 0x1;
      } else if (((mode == SET_DAY) && (region == REGION_US)) ||
	((mode == SET_MONTH) && (region == REGION_EU))) {
	mode = SET_YEAR;
	display_date(DATE);
	display[7] |= 0x1;
	display[8] |= 0x1;
      } else {
	displaymode = NONE;
	display_date(DATE);
	delayms(1500);
	displaymode = SHOW_TIME;
	return;
      }
    }
    if ((just_pressed & 0x4) || (pressed & 0x4)) {
      just_pressed = 0;
      if (mode == SET_MONTH) {
	date_m++;
	if (date_m >= 13)
	  date_m = 1;
	display_date(DATE);
	if (region == REGION_US) {
	  display[1] |= 0x1;
	  display[2] |= 0x1;
	} else {
	  display[4] |= 0x1;
	  display[5] |= 0x1;
	}
	eeprom_write_byte((uint8_t *)EE_MONTH, date_m);    
      }
      if (mode == SET_DAY) {
	date_d++;
	if (date_d > 31)
	  date_d = 1;
	display_date(DATE);

	if (region == REGION_EU) {
	  display[1] |= 0x1;
	  display[2] |= 0x1;
	} else {
	  display[4] |= 0x1;
	  display[5] |= 0x1;
	}
	eeprom_write_byte((uint8_t *)EE_DAY, date_d);    
      }
      if (mode == SET_YEAR) {
	date_y++;
	date_y %= 100;
	display_date(DATE);
	display[7] |= 0x1;
	display[8] |= 0x1;
	eeprom_write_byte((uint8_t *)EE_YEAR, date_y);    
      }

      if (pressed & 0x4)
	delayms(60);
    }
  }
}

void set_time(void) 
{
  uint8_t mode;
  uint8_t hour, min, sec;
    
  hour = min = sec = 0;
  mode = SHOW_MENU;

  timeoutcounter = INACTIVITYTIMEOUT;
  
  while (1) {
    if (just_pressed & 0x1) { // mode change
      return;
    }
    if (just_pressed || pressed) {
      timeoutcounter = INACTIVITYTIMEOUT;  
      // timeout w/no buttons pressed after 3 seconds?
    } else if (!timeoutcounter) {
      //timed out!
      time_h = hour;
      time_m = min;
      time_s = sec;
      displaymode = SHOW_TIME;     
      return;
    }
    if (just_pressed & 0x2) {
      just_pressed = 0;
      if (mode == SHOW_MENU) {
	hour = time_h;
	min = time_m;
	sec = time_s;

	// ok now its selected
	mode = SET_HOUR;
	display_time(hour, min, sec);
	display[1] |= 0x1;
	display[2] |= 0x1;	
      } else if (mode == SET_HOUR) {
	mode = SET_MIN;
	display_time(hour, min, sec);
	display[4] |= 0x1;
	display[5] |= 0x1;
      } else if (mode == SET_MIN) {
	mode = SET_SEC;
	display_time(hour, min, sec);
	display[7] |= 0x1;
	display[8] |= 0x1;
      } else {
	// done!
	time_h = hour;
	time_m = min;
	time_s = sec;
	displaymode = SHOW_TIME;
	return;
      }
    }
    if ((just_pressed & 0x4) || (pressed & 0x4)) {
      just_pressed = 0;
      
      if (mode == SET_HOUR) {
	hour = (hour+1) % 24;
	display_time(hour, min, sec);
	display[1] |= 0x1;
	display[2] |= 0x1;
	time_h = hour;
	eeprom_write_byte((uint8_t *)EE_HOUR, time_h);    
      }
      if (mode == SET_MIN) {
	min = (min+1) % 60;
	display_time(hour, min, sec);
	display[4] |= 0x1;
	display[5] |= 0x1;
	eeprom_write_byte((uint8_t *)EE_MIN, time_m);
	time_m = min;
      }
      if ((mode == SET_SEC) ) {
	sec = (sec+1) % 60;
	display_time(hour, min, sec);
	display[7] |= 0x1;
	display[8] |= 0x1;
	time_s = sec;
      }
      
      if (pressed & 0x4)
	delayms(75);
    }
  }
}


void set_alarm(void) 
{
  uint8_t mode;
  uint8_t hour, min, sec;
    
  hour = min = sec = 0;
  mode = SHOW_MENU;

  timeoutcounter = INACTIVITYTIMEOUT;
  
  while (1) {
    if (just_pressed & 0x1) { // mode change
      return;
    }
   if (just_pressed || pressed) {
      timeoutcounter = INACTIVITYTIMEOUT;  
      // timeout w/no buttons pressed after 3 seconds?
    } else if (!timeoutcounter) {
      //timed out!
      displaymode = SHOW_TIME;     
      return;
    }
    if (just_pressed & 0x2) {
      just_pressed = 0;
      if (mode == SHOW_MENU) {
	hour = alarm_h;
	min = alarm_m;
	sec = 0;

	// ok now its selected
	mode = SET_HOUR;
	display_alarm(hour, min);
	display[1] |= 0x1;
	display[2] |= 0x1;	
      } else if (mode == SET_HOUR) {
	mode = SET_MIN;
	display_alarm(hour, min);
	display[4] |= 0x1;
	display[5] |= 0x1;
      } else {
	// done!
	displaymode = SHOW_TIME;
	return;
      }
    }
    if ((just_pressed & 0x4) || (pressed & 0x4)) {
      just_pressed = 0;

      if (mode == SET_HOUR) {
	hour = (hour+1) % 24;
	display_alarm(hour, min);
	display[1] |= 0x1;
	display[2] |= 0x1;
	alarm_h = hour;
	eeprom_write_byte((uint8_t *)EE_ALARM_HOUR, alarm_h);    
      }
      if (mode == SET_MIN) {
	min = (min+1) % 60;
	display_alarm(hour, min);
	display[4] |= 0x1;
	display[5] |= 0x1;
	alarm_m = min;
	eeprom_write_byte((uint8_t *)EE_ALARM_MIN, alarm_m);    
      }

      if (pressed & 0x4)
	delayms(75);
    }
  }
}



/**************************** RTC & ALARM *****************************/
void clock_init(void) {
  
  time_h = eeprom_read_byte((uint8_t *)EE_HOUR) % 24;
  time_m = eeprom_read_byte((uint8_t *)EE_MIN) % 60;
  time_s = eeprom_read_byte((uint8_t *)EE_SEC) % 60;
  /*
  time_h = TIMEHOUR;
  time_m = TIMEMIN;
  time_s = TIMESEC + 10;
  */
  alarm_m = eeprom_read_byte((uint8_t *)EE_ALARM_MIN) % 60;
  alarm_h = eeprom_read_byte((uint8_t *)EE_ALARM_HOUR) % 24;

  date_y = eeprom_read_byte((uint8_t *)EE_YEAR) % 100;
  date_m = eeprom_read_byte((uint8_t *)EE_MONTH) % 13;
  date_d = eeprom_read_byte((uint8_t *)EE_DAY) % 32;

  ASSR |= _BV(AS2); // use crystal
  TCCR2B = _BV(CS22) | _BV(CS20); // div by 128
  // overflow once a second

  // enable interrupt
  TIMSK2 = _BV(TOIE2);

  sei();
}


void setalarmstate(void) {
  if (ALARM_PIN & _BV(ALARM)) { 
    if  (!alarm_on) {
      // alarm on!
      alarm_on = 1;
      snoozetimer = 0;
      display_str("alarm on");
      displaymode = SHOW_SNOOZE;
      beep(4000, 1);
      delayms(1000);
      display_alarm(alarm_h, alarm_m);
      delayms(1000);
      displaymode = SHOW_TIME;
    }
  } else {
    if (alarm_on) {
      alarm_on = 0;
      snoozetimer = 0;
      if (alarming) {
	DEBUGP("alarm off");
	alarming = 0;
	TCCR1B &= ~_BV(CS11); // turn it off!
	PORTB |= _BV(SPK1) | _BV(SPK2);
      } else {
	beep(1000, 1);
      }
    }
  }
}

uint8_t leapyear(uint16_t y) {
  return ( (!(y % 4) && (y % 100)) || !(y % 400));
}


/**************************** SPEAKER *****************************/
void speaker_init(void) {
  volume = eeprom_read_byte((uint8_t *)EE_VOLUME);

    // fast PWM, 8 bit
  PORTB |= _BV(SPK1) | _BV(SPK2); 

  TCCR1A = _BV(COM1B1) | _BV(COM1B0) | _BV(WGM11);
  if (volume) {
    TCCR1A |= _BV(COM1A1);
  } 
  TCCR1B = _BV(WGM13) | _BV(WGM12);
 // start at 4khz
  ICR1 = 250;
  OCR1B = OCR1A = ICR1 / 2;
}

void tick(void) {
  TCCR1A = 0;
  TCCR1B = 0;

  SPK_PORT |= _BV(SPK1);
  SPK_PORT &= ~_BV(SPK2);
  delayms(10);
  SPK_PORT |= _BV(SPK2);
  SPK_PORT &= ~_BV(SPK1);
  delayms(10);
  SPK_PORT &= ~_BV(SPK1) & ~_BV(SPK2);

  TCCR1A = _BV(COM1A1) | _BV(COM1B1) | _BV(COM1B0) | _BV(WGM11);
  TCCR1B = _BV(WGM13) | _BV(WGM12);
}

void beep(uint16_t freq, uint8_t times) {
  ICR1 = (F_CPU/8)/freq;
  OCR1A = OCR1B = ICR1/2;
   
  while (times--) {
    TCCR1B |= _BV(CS11); // turn it on!
    _delay_ms(200);
    TCCR1B &= ~_BV(CS11); // turn it off!
    PORTB &= ~_BV(SPK1) & ~_BV(SPK2);
    _delay_ms(200);
  }
  PORTB &= ~_BV(SPK1) & ~_BV(SPK2);
}


/**************************** DISPLAY *****************************/


void display_date(uint8_t style) {
  if (style == DATE) {
    display[0] = 0;
    display[6] = display[3] = 0x02;
    if (region == REGION_US) {
      display[1] = pgm_read_byte(numbertable_p + (date_m / 10));
      display[2] = pgm_read_byte(numbertable_p + (date_m % 10));
      display[4] = pgm_read_byte(numbertable_p + (date_d / 10));
      display[5] = pgm_read_byte(numbertable_p + (date_d % 10));
    } else {
      display[1] = pgm_read_byte(numbertable_p + (date_d / 10));
      display[2] = pgm_read_byte(numbertable_p + (date_d % 10));
      display[4] = pgm_read_byte(numbertable_p + (date_m / 10));
      display[5] = pgm_read_byte(numbertable_p + (date_m % 10));
    }
    display[7] = pgm_read_byte(numbertable_p + (date_y / 10));
    display[8] = pgm_read_byte(numbertable_p + (date_y % 10));
  } else if (style == DAY) {
    // calculate day of the week
    uint16_t month, year;
    uint8_t dotw;
    
    month = date_m;
    year = 2000 + date_y;
    if (date_m < 3)  {
      month += 12;
      year -= 1;
    }
    dotw = (date_d + (2 * month) + (6 * (month+1)/10) + year + (year/4) - (year/100) + (year/400) + 1) % 7;

    display[8] = display[7] = 0;
    switch (dotw) {
    case 0:
      display_str("sunday"); break;
    case 1:
      display_str("monday"); break;
    case 2:
      display_str("tuesday"); break;
    case 3:
      display_str("wednsday"); break;
    case 4:
      display_str("thursday"); break;
    case 5:
      display_str("friday"); break;
    case 6:
      display_str("saturday"); break;
    }
    
    delayms(1000);

    display[6] = display[5] = display[4] = 0;
    switch (date_m) {
    case 1:
      display_str("jan"); break;
    case 2:
      display_str("feb"); break;
    case 3:
      display_str("march"); break;
    case 4:
      display_str("april"); break;
    case 5:
      display_str("may"); break;
    case 6:
      display_str("june"); break;
    case 7:
      display_str("july"); break;
    case 8:
      display_str("augst"); break;
    case 9:
      display_str("sept"); break;
    case 10:
      display_str("octob"); break;
    case 11:
      display_str("novem"); break;
    case 12:
      display_str("decem"); break;
    }
    display[7] = pgm_read_byte(numbertable_p + (date_d / 10));
    display[8] = pgm_read_byte(numbertable_p + (date_d % 10));
    
  }
}

void display_time(uint8_t h, uint8_t m, uint8_t s) {
  display[8] =  pgm_read_byte(numbertable_p + (s % 10));
  display[7] =  pgm_read_byte(numbertable_p + (s / 10));
  display[6] = 0;
  display[5] =  pgm_read_byte(numbertable_p + (m % 10));
  display[4] =  pgm_read_byte(numbertable_p + (m / 10)); 
  display[3] = 0;

  // check euro or US style time
  if (region == REGION_US) {
    if (h >= 12)
      display[0] |= 0x1;  // 'pm' notice
    else 
      display[0] &= ~0x1;  // 'pm' notice
    display[2] =  pgm_read_byte(numbertable_p + ( (((h+11)%12)+1) % 10));
    display[1] =  pgm_read_byte(numbertable_p + ( (((h+11)%12)+1) / 10));
  } else {
    display[2] =  pgm_read_byte(numbertable_p + ( (((h+23)%24)+1) % 10));
    display[1] =  pgm_read_byte(numbertable_p + ( (((h+23)%24)+1) / 10));
  }
}

void display_alarm(uint8_t h, uint8_t m){ 
  display[8] = 0;
  display[7] = 0;
  display[6] = 0;
  display[5] = pgm_read_byte(numbertable_p + (m % 10));
  display[4] = pgm_read_byte(numbertable_p + (m / 10)); 
  display[3] = 0;

  // check euro or US style time
  if (region == REGION_US) {
    if (h >= 12) {
      display[0] |= 0x1;  // 'pm' notice
      display[7] = pgm_read_byte(alphatable_p + 'p' - 'a');
    } else {
      display[7] = pgm_read_byte(alphatable_p + 'a' - 'a');
      display[0] &= ~0x1;  // 'am' notice
    }
    display[8] = pgm_read_byte(alphatable_p + 'm' - 'a');

    display[2] =  pgm_read_byte(numbertable_p + ( (((h+11)%12)+1) % 10));
    display[1] =  pgm_read_byte(numbertable_p + ( (((h+11)%12)+1) / 10));
  } else {
      display[2] =  pgm_read_byte(numbertable_p + ( (((h+23)%24)+1) % 10));
    display[1] =  pgm_read_byte(numbertable_p + ( (((h+23)%24)+1) / 10));
  }
}


// display words (menus, prompts, etc)
void display_str(char *s) {
  uint8_t i;

  display[0] = 0;
  for (i=1; i<9; i++) {
    if (s[i-1] == 0)
      return;
    if ((s[i-1] >= 'a') && (s[i-1] <= 'z')) {
      display[i] =  pgm_read_byte(alphatable_p + s[i-1] - 'a');
    } else if ((s[i-1] >= '0') && (s[i-1] <= '9')) {
      display[i] =  pgm_read_byte(numbertable_p + s[i-1] - '0');
    } else {
      display[i] = 0;
    }
  }
}

// setup SPI
void vfd_init(void) {
  SPCR  = _BV(SPE) | _BV(MSTR) | _BV(SPR0);
}

void boost_init(uint8_t pwm) {
  // fast PWM, set OC0A on match
  TCCR0A = _BV(WGM00) | _BV(WGM01);  

  // fastest clock
  TCCR0B = _BV(CS00);

  // set PWM value
  if (pwm > 90)
    pwm = 90;
  if (pwm < 10)
    pwm = 10;
  OCR0A = pwm;

  TCCR0A |= _BV(COM0A1);
  TIMSK0 |= _BV(OCIE0A); // turn on the interrupt for muxing
  sei();
}



// set a particular display digit and then update display
void setdisplay(uint8_t digit, uint8_t segments) {
  uint32_t d = 0;
  uint8_t i;

  // set the digit
  d |= _BV(pgm_read_byte(digittable_p + digit));

  
  // set the segments
  for (i=0; i<8; i++) {
    if (segments & _BV(i)) {
      t = 1;
      t <<= pgm_read_byte(segmenttable_p + i);
      d |= t;
    }
  }
  vfd_send(d);
}

// send raw data to display
void vfd_send(uint32_t d) {
  // send lowest 20 bits
  cli();
  spi_xfer(d >> 16);
  spi_xfer(d >> 8);
  spi_xfer(d);
  // latch data

  VFDLOAD_PORT |= _BV(VFDLOAD);
  VFDLOAD_PORT &= ~_BV(VFDLOAD);
  sei();
}

// send 1 byte
void spi_xfer(uint8_t c) {

  SPDR = c;
  while (! (SPSR & _BV(SPIF)));
}

